import torch
import random

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from PIL import Image
from tqdm import tqdm
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from cftool.misc import safe_execute
from cftool.misc import shallow_copy_dict
from cftool.misc import context_error_handler
from cftool.array import arr_type
from cftool.array import save_images
from cftool.types import tensor_dict_type

from .common import read_image
from .common import restrict_wh
from .common import get_suitable_size
from .common import APIMixin
from .common import ReadImageResponse
from ....zoo import DLZoo
from ....data import predict_tensor_data
from ....data import TensorInferenceData
from ....pipeline import DLPipeline
from ....constants import INPUT_KEY
from ....misc.toolkit import slerp
from ....misc.toolkit import new_seed
from ....misc.toolkit import eval_context
from ....misc.toolkit import seed_everything
from ....modules.blocks import Conv2d
from ....models.cv.diffusion import LDM
from ....models.cv.diffusion import DDPM
from ....models.cv.diffusion import ISampler
from ....models.cv.ae.common import IAutoEncoder
from ....models.cv.diffusion.utils import get_timesteps
from ....models.cv.diffusion.utils import CONCAT_KEY
from ....models.cv.diffusion.utils import CONCAT_TYPE
from ....models.cv.diffusion.utils import HYBRID_TYPE
from ....models.cv.diffusion.utils import CROSS_ATTN_KEY
from ....models.cv.diffusion.cond_models import CLIPTextConditionModel
from ....models.cv.diffusion.samplers.ddim import DDIMMixin
from ....models.cv.diffusion.samplers.solver import DPMSolver
from ....models.cv.diffusion.samplers.k_samplers import KSamplerMixin


class switch_sampler_context:
    def __init__(self, api: "DiffusionAPI", sampler: Optional[str]):
        self.api = api
        self.m_sampler = api.sampler.__identifier__
        self.target_sampler = sampler
        self.switched = False

    def __enter__(self) -> None:
        if self.target_sampler is None:
            return
        if self.m_sampler != self.target_sampler:
            self.switched = True
            self.api.switch_sampler(self.target_sampler)

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self.switched:
            self.api.switch_sampler(self.m_sampler)


class SizeInfo(NamedTuple):
    factor: int
    opt_size: int


class MaskedCond(NamedTuple):
    image_res: ReadImageResponse
    mask_res: ReadImageResponse
    mask: np.ndarray
    remained_image: np.ndarray
    remained_mask: np.ndarray
    mask_cond: Tensor
    remained_image_cond: Tensor


class DiffusionAPI(APIMixin):
    m: DDPM
    sampler: ISampler
    cond_model: Optional[nn.Module]
    first_stage: Optional[IAutoEncoder]
    latest_seed: int
    latest_variation_seed: Optional[int]

    def __init__(
        self,
        m: DDPM,
        device: torch.device,
        *,
        use_amp: bool = False,
        use_half: bool = False,
        clip_skip: int = 0,
    ):
        super().__init__(m, device, use_amp=use_amp, use_half=use_half)
        self.sampler = m.sampler
        self.cond_type = m.condition_type
        self.clip_skip = clip_skip
        # extracted the condition model so we can pre-calculate the conditions
        self.cond_model = m.condition_model
        if self.cond_model is not None:
            self.cond_model.eval()
        m.condition_model = nn.Identity()
        # pre-calculate unconditional_cond if needed
        self._original_raw_uncond = getattr(m.sampler, "unconditional_cond", None)
        self._uncond_cache: tensor_dict_type = {}
        self._update_sampler_uncond(clip_skip)
        # extract first stage
        if not isinstance(m, LDM):
            self.first_stage = None
        else:
            self.first_stage = m.first_stage.core

    def to(
        self,
        device: torch.device,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> None:
        super().to(device, use_amp=use_amp, use_half=use_half)
        unconditional_cond = getattr(self.sampler, "unconditional_cond", None)
        if use_half:
            if self.cond_model is not None:
                self.cond_model.half()
            if self.first_stage is not None:
                self.first_stage.half()
            if unconditional_cond is not None:
                unconditional_cond = unconditional_cond.half()
            for k, v in self._uncond_cache.items():
                self._uncond_cache[k] = v.half()
        else:
            if self.cond_model is not None:
                self.cond_model.float()
            if self.first_stage is not None:
                self.first_stage.float()
            if unconditional_cond is not None:
                unconditional_cond = unconditional_cond.float()
            for k, v in self._uncond_cache.items():
                self._uncond_cache[k] = v.float()
        if self.cond_model is not None:
            self.cond_model.to(device)
        if self.first_stage is not None:
            self.first_stage.to(device)
        if unconditional_cond is not None:
            self.sampler.unconditional_cond = unconditional_cond.to(device)
        for k, v in self._uncond_cache.items():
            self._uncond_cache[k] = v.to(device)

    @property
    def size_info(self) -> SizeInfo:
        opt_size = self.m.img_size
        if self.first_stage is None:
            factor = 1
        else:
            factor = self.first_stage.img_size // opt_size
        return SizeInfo(factor, opt_size)

    def get_cond(self, cond: Any) -> Tensor:
        if self.cond_model is None:
            msg = "should not call `get_cond` when `cond_model` is not available"
            raise ValueError(msg)
        with torch.no_grad():
            with self.amp_context:
                return self.cond_model(cond)

    def switch_sampler(
        self,
        sampler: str,
        sampler_config: Optional[Dict[str, Any]] = None,
    ) -> None:
        sampler_ins = self.m.make_sampler(sampler, sampler_config)
        current_unconditional_cond = getattr(self.m.sampler, "unconditional_cond", None)
        if current_unconditional_cond is not None:
            if hasattr(sampler_ins, "unconditional_cond"):
                sampler_ins.unconditional_cond = current_unconditional_cond
        current_guidance = getattr(self.m.sampler, "unconditional_guidance_scale", None)
        if current_guidance is not None:
            if hasattr(sampler_ins, "unconditional_guidance_scale"):
                sampler_ins.unconditional_guidance_scale = current_guidance
        self.sampler = self.m.sampler = sampler_ins

    def switch_circular(self, enable: bool) -> None:
        def _inject(m: nn.Module) -> None:
            for child in m.children():
                _inject(child)
            modules.append(m)

        padding_mode = "circular" if enable else "zeros"
        modules: List[nn.Module] = []
        _inject(self.m)
        for module in modules:
            if isinstance(module, nn.Conv2d):
                module.padding_mode = padding_mode
            elif isinstance(module, Conv2d):
                module.padding = padding_mode

    def sample(
        self,
        num_samples: int,
        export_path: Optional[str] = None,
        *,
        seed: Optional[int] = None,
        use_seed_resize: bool = False,
        # each variation contains (seed, weight)
        variations: Optional[List[Tuple[int, float]]] = None,
        variation_seed: Optional[int] = None,
        variation_strength: Optional[float] = None,
        z: Optional[Tensor] = None,
        z_ref: Optional[Tensor] = None,
        z_ref_mask: Optional[Tensor] = None,
        z_ref_noise: Optional[Tensor] = None,
        size: Optional[Tuple[int, int]] = None,
        original_size: Optional[Tuple[int, int]] = None,
        alpha: Optional[np.ndarray] = None,
        cond: Optional[Any] = None,
        cond_concat: Optional[Tensor] = None,
        unconditional_cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        callback: Optional[Callable[[Tensor], Tensor]] = None,
        batch_size: int = 1,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        registered_custom = False
        if self.cond_model is not None:
            clip_skip = kwargs.get(
                "clip_skip",
                0 if self.clip_skip is None else self.clip_skip,
            )
            self._update_sampler_uncond(clip_skip)
            if isinstance(self.cond_model, CLIPTextConditionModel):
                custom_embeddings = kwargs.get("custom_embeddings")
                if custom_embeddings is not None:
                    registered_custom = True
                    self.cond_model.register_custom(custom_embeddings)
        if cond is not None:
            if self.cond_type != CONCAT_TYPE and self.cond_model is not None:
                data = TensorInferenceData(cond, batch_size=batch_size)
                cond = predict_tensor_data(self.cond_model, data)
        if cond is not None and num_samples != len(cond):
            raise ValueError(
                f"`num_samples` ({num_samples}) should be identical with "
                f"the number of `cond` ({len(cond)})"
            )
        if alpha is not None and original_size is not None:
            alpha_h, alpha_w = alpha.shape[-2:]
            if alpha_w != original_size[0] or alpha_h != original_size[1]:
                raise ValueError(
                    f"shape of the provided `alpha` ({alpha_w}, {alpha_h}) should be "
                    f"identical with the provided `original_size` {original_size}"
                )
        unconditional = cond is None
        if unconditional:
            cond = [0] * num_samples
        iterator = TensorInferenceData(cond, batch_size=batch_size).initialize()[0]
        num_iter = len(iterator)
        if verbose and num_iter > 1:
            iterator = tqdm(iterator, desc="iter", total=num_iter)
        sampled = []
        kw = dict(num_steps=num_steps, verbose=verbose)
        kw.update(shallow_copy_dict(kwargs))
        factor, opt_size = self.size_info
        if size is None:
            size = opt_size, opt_size
        else:
            size = tuple(map(lambda n: round(n / factor), size))  # type: ignore
        uncond_backup = None
        unconditional_cond_backup = None
        if self.cond_model is not None and unconditional_cond is not None:
            uncond_backup = getattr(self.sampler, "uncond", None)
            unconditional_cond_backup = getattr(
                self.sampler,
                "unconditional_cond",
                None,
            )
            uncond = self.get_cond(unconditional_cond).to(self.device)
            self.sampler.uncond = uncond.clone()
            self.sampler.unconditional_cond = uncond.clone()
        with eval_context(self.m):
            with self.amp_context:
                for batch in iterator:
                    i_kw = shallow_copy_dict(kw)
                    i_cond = batch[INPUT_KEY].to(self.device)
                    repeat = lambda t: t.repeat_interleave(len(i_cond), dim=0)
                    if z is not None:
                        i_z = repeat(z)
                    else:
                        in_channels = self.m.in_channels
                        if self.cond_type == CONCAT_TYPE:
                            in_channels -= cond.shape[1]
                        elif cond_concat is not None:
                            in_channels -= cond_concat.shape[1]
                        i_z_shape = len(i_cond), in_channels, *size[::-1]
                        i_z, _ = self._set_seed_and_variations(
                            seed,
                            lambda: torch.randn(i_z_shape, device=self.device),
                            lambda noise: noise,
                            variations,
                            variation_seed,
                            variation_strength,
                        )
                    if use_seed_resize:
                        z_original_shape = list(i_z.shape[-2:])
                        z_opt_shape = list(
                            map(lambda n: round(n / factor), [opt_size, opt_size])
                        )
                        if z_original_shape != z_opt_shape:
                            dx = (z_original_shape[0] - z_opt_shape[0]) // 2
                            dy = (z_original_shape[1] - z_opt_shape[1]) // 2
                            x = z_opt_shape[0] if dx >= 0 else z_opt_shape[0] + 2 * dx
                            y = z_opt_shape[1] if dy >= 0 else z_opt_shape[1] + 2 * dy
                            dx = max(-dx, 0)
                            dy = max(-dy, 0)
                            i_opt_z_shape = (
                                len(i_cond),
                                self.m.in_channels,
                                *z_opt_shape,
                            )
                            i_opt_z, _ = self._set_seed_and_variations(
                                seed,
                                lambda: torch.randn(i_opt_z_shape, device=self.device),
                                lambda noise: noise,
                                variations,
                                variation_seed,
                                variation_strength,
                            )
                            i_z[..., dx : dx + x, dy : dy + y] = i_opt_z[
                                ..., dx : dx + x, dy : dy + y
                            ]
                    if z_ref is not None and z_ref_mask is not None:
                        if z_ref_noise is not None:
                            i_kw["ref"] = repeat(z_ref)
                            i_kw["ref_mask"] = repeat(z_ref_mask)
                            i_kw["ref_noise"] = repeat(z_ref_noise)
                    if unconditional:
                        i_cond = None
                    if self.use_half:
                        i_z = i_z.half()
                        if i_cond is not None:
                            i_cond = i_cond.half()
                        for k, v in i_kw.items():
                            if isinstance(v, torch.Tensor) and v.is_floating_point():
                                i_kw[k] = v.half()
                    if cond_concat is not None:
                        if self.cond_type != HYBRID_TYPE:
                            raise ValueError(
                                f"condition type should be `{HYBRID_TYPE}` when "
                                f"`cond_concat` is provided"
                            )
                        i_cond = {
                            CROSS_ATTN_KEY: i_cond,
                            CONCAT_KEY: cond_concat,
                        }
                    with switch_sampler_context(self, i_kw.get("sampler")):
                        i_sampled = self.m.decode(i_z, cond=i_cond, **i_kw)
                    sampled.append(i_sampled.cpu().float())
        if uncond_backup is not None:
            self.sampler.uncond = uncond_backup
        if unconditional_cond_backup is not None:
            self.sampler.unconditional_cond = unconditional_cond_backup
        concat = torch.cat(sampled, dim=0)
        if clip_output:
            concat = torch.clip(concat, -1.0, 1.0)
        if callback is not None:
            concat = callback(concat)
        if original_size is not None:
            original_size = (max(original_size[0], 1), max(original_size[1], 1))
            with torch.no_grad():
                concat = F.interpolate(
                    concat,
                    original_size[::-1],
                    mode="bicubic",
                )
        if alpha is not None:
            alpha = torch.from_numpy(2.0 * alpha - 1.0)
            if original_size is None:
                with torch.no_grad():
                    alpha = F.interpolate(
                        alpha,
                        concat.shape[-2:],
                        mode="nearest",
                    )
            concat = torch.cat([concat, alpha], dim=1)
        if export_path is not None:
            save_images(concat, export_path)
        self.empty_cuda_cache()
        if registered_custom:
            self.cond_model.clear_custom()
        return concat

    def _update_clip_skip(self, clip_skip: int) -> None:
        if isinstance(self.cond_model, CLIPTextConditionModel):
            self.cond_model.clip_skip = clip_skip

    def _update_sampler_uncond(self, clip_skip: int) -> None:
        self._update_clip_skip(clip_skip)
        if self.cond_model is not None and self._original_raw_uncond is not None:
            cache = self._uncond_cache.get(clip_skip)
            if cache is not None:
                uncond = cache
            else:
                uncond = self.get_cond(self._original_raw_uncond)
                self._uncond_cache[clip_skip] = uncond
            self.m.sampler.unconditional_cond = uncond.to(self.device)

    @staticmethod
    def _txt_cond(t: Union[str, List[str]], n: Optional[int]) -> Tuple[List[str], int]:
        if n is None:
            n = 1 if isinstance(t, str) else len(t)
        if isinstance(t, str):
            t = [t] * n
        if len(t) != n:
            raise ValueError(
                f"`num_samples` ({n}) should be identical with "
                f"the number of `txt` ({len(t)})"
            )
        return t, n

    @staticmethod
    def _get_size(
        size: Optional[Tuple[int, int]],
        anchor: int,
        max_wh: int,
    ) -> Optional[Tuple[int, int]]:
        if size is None:
            return None
        new_size = restrict_wh(*size, max_wh)
        return tuple(map(get_suitable_size, new_size, (anchor, anchor)))  # type: ignore

    def _get_masked_cond(
        self,
        image: Union[str, Image.Image],
        mask: Union[str, Image.Image],
        max_wh: int,
        anchor: int,
        mask_image_fn: Callable[[np.ndarray, np.ndarray], np.ndarray],
        mask_cond_fn: Callable[[np.ndarray], Tensor],
    ) -> MaskedCond:
        # handle mask stuffs
        image_res = read_image(image, max_wh, anchor=anchor)
        mask_res = read_image(mask, max_wh, anchor=anchor, to_mask=True)
        mask = mask_res.image
        bool_mask = mask >= 0.5
        remained_mask = (~bool_mask).astype(np.float16 if self.use_half else np.float32)
        remained_image = mask_image_fn(remained_mask, image_res.image)
        # construct condition tensor
        remained_cond = self._get_z(remained_image)
        latent_shape = remained_cond.shape[-2:]
        mask_cond = mask_cond_fn(bool_mask)
        mask_cond = mask_cond.to(torch.float16 if self.use_half else torch.float32)
        mask_cond = mask_cond.to(self.device)
        mask_cond = F.interpolate(mask_cond, size=latent_shape)
        return MaskedCond(
            image_res,
            mask_res,
            mask,
            remained_image,
            remained_mask,
            mask_cond,
            remained_cond,
        )

    def txt2img(
        self,
        txt: Union[str, List[str]],
        export_path: Optional[str] = None,
        *,
        anchor: int = 64,
        max_wh: int = 512,
        num_samples: Optional[int] = None,
        size: Optional[Tuple[int, int]] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        callback: Optional[Callable[[Tensor], Tensor]] = None,
        batch_size: int = 1,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        txt, num_samples = self._txt_cond(txt, num_samples)
        new_size = self._get_size(size, anchor, max_wh)
        return self.sample(
            num_samples,
            export_path,
            size=new_size,
            original_size=size,
            cond=txt,
            num_steps=num_steps,
            clip_output=clip_output,
            callback=callback,
            batch_size=batch_size,
            verbose=verbose,
            **kwargs,
        )

    def txt2img_inpainting(
        self,
        txt: str,
        image: Union[str, Image.Image],
        mask: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        reference: Optional[Union[str, Image.Image]] = None,
        reference_fidelity: float = 0.2,
        anchor: int = 64,
        max_wh: int = 512,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        keep_original: bool = False,
        use_raw_inpainting: bool = False,
        raw_inpainting_fidelity: float = 0.2,
        callback: Optional[Callable[[Tensor], Tensor]] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        if use_raw_inpainting:
            image_res = read_image(image, max_wh, anchor=anchor)
            mask = read_image(mask, max_wh, anchor=anchor, to_mask=True).image
            z = z_ref = self._get_z(image_res.image)
            z_ref_mask = F.interpolate(
                torch.from_numpy(mask).to(z_ref),
                z_ref.shape[-2:],
                mode="bicubic",
            )
            return self._img2img(
                z,
                export_path,
                z_ref=z_ref,
                z_ref_mask=z_ref_mask,
                fidelity=raw_inpainting_fidelity,
                original_size=image_res.original_size,
                alpha=None,
                cond=[txt],
                num_steps=num_steps,
                clip_output=clip_output,
                verbose=verbose,
                **kwargs,
            )
        txt_list, num_samples = self._txt_cond(txt, 1)
        res = self._get_masked_cond(
            image,
            mask,
            max_wh,
            anchor,
            lambda remained_mask, img: np.where(remained_mask, img, 0.5),
            lambda bool_mask: torch.from_numpy(bool_mask),
        )
        # sampling
        with switch_sampler_context(self, kwargs.get("sampler")):
            if reference is None:
                z = None
                size = tuple(
                    map(
                        lambda n: n * self.size_info.factor,
                        res.remained_image_cond.shape[-2:][::-1],
                    )
                )
            else:
                size = None
                z = self._get_z(read_image(reference, max_wh, anchor=anchor).image)
                z, _, kwargs = self._q_sample(
                    z,
                    num_steps,
                    reference_fidelity,
                    **kwargs,
                )
            sampled = self.sample(
                num_samples,
                export_path,
                z=z,
                size=size,  # type: ignore
                original_size=res.image_res.original_size,
                alpha=res.image_res.alpha,
                cond=txt_list,
                cond_concat=torch.cat([res.mask_cond, res.remained_image_cond], dim=1),
                num_steps=num_steps,
                clip_output=clip_output,
                callback=callback,
                verbose=verbose,
                **kwargs,
            )
        if keep_original:
            original = np.array(res.image_res.original).astype(np.float32) / 127.5 - 1.0
            original = original.transpose([2, 0, 1])[None]
            mask_ = read_image(res.mask_res.original, None, anchor=None, to_mask=True)
            remained_mask = ~(mask_.image >= 0.5)
            sampled = np.where(remained_mask, original, sampled.numpy())
            sampled = torch.from_numpy(sampled)
        return sampled

    def outpainting(
        self,
        txt: str,
        image: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 64,
        max_wh: int = 512,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        keep_original: bool = False,
        callback: Optional[Callable[[Tensor], Tensor]] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        if isinstance(image, str):
            image = Image.open(image)
        if image.mode != "RGBA":
            raise ValueError("`image` should be `RGBA` in outpainting")
        *rgb, alpha = image.split()
        mask = Image.fromarray(255 - np.array(alpha))
        image = Image.merge("RGB", rgb)
        return self.txt2img_inpainting(
            txt,
            image,
            mask,
            export_path,
            anchor=anchor,
            max_wh=max_wh,
            num_steps=num_steps,
            clip_output=clip_output,
            keep_original=keep_original,
            callback=callback,
            verbose=verbose,
            **kwargs,
        )

    def img2img(
        self,
        image: Union[str, Image.Image, Tensor],
        export_path: Optional[str] = None,
        *,
        anchor: int = 32,
        max_wh: int = 512,
        fidelity: float = 0.2,
        alpha: Optional[np.ndarray] = None,
        cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        if isinstance(image, Tensor):
            original_size = tuple(image.shape[-2:][::-1])
        else:
            res = read_image(image, max_wh, anchor=anchor)
            image = res.image
            original_size = res.original_size
            if alpha is None:
                alpha = res.alpha
        z = self._get_z(image)
        return self._img2img(
            z,
            export_path,
            fidelity=fidelity,
            original_size=original_size,  # type: ignore
            alpha=alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            verbose=verbose,
            **kwargs,
        )

    def inpainting(
        self,
        image: Union[str, Image.Image],
        mask: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 32,
        max_wh: int = 512,
        alpha: Optional[np.ndarray] = None,
        refine_fidelity: Optional[float] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        # inpainting callback, will not trigger in refine stage
        def callback(out: Tensor) -> Tensor:
            final = torch.from_numpy(res.remained_image.copy())
            final += 0.5 * (1.0 + out) * (1.0 - res.remained_mask)
            return 2.0 * final - 1.0

        res = self._get_masked_cond(
            image,
            mask,
            max_wh,
            anchor,
            lambda remained_mask, img: np.where(remained_mask, img, 0.0),
            lambda bool_mask: torch.where(torch.from_numpy(bool_mask), 1.0, -1.0),
        )
        cond = torch.cat([res.remained_image_cond, res.mask_cond], dim=1)
        size = tuple(
            map(
                lambda n: n * self.size_info.factor,
                res.remained_image_cond.shape[-2:][::-1],
            )
        )
        # refine with img2img
        if refine_fidelity is not None:
            z = self._get_z(res.image_res.image)
            return self._img2img(
                z,
                export_path,
                fidelity=refine_fidelity,
                original_size=res.image_res.original_size,
                alpha=res.image_res.alpha if alpha is None else alpha,
                cond=cond,
                num_steps=num_steps,
                clip_output=clip_output,
                verbose=verbose,
                **kwargs,
            )
        # sampling
        return self.sample(
            1,
            export_path,
            size=size,  # type: ignore
            original_size=res.image_res.original_size,
            alpha=res.image_res.alpha if alpha is None else alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            callback=callback,
            verbose=verbose,
            **kwargs,
        )

    def sr(
        self,
        image: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 8,
        max_wh: int = 512,
        alpha: Optional[np.ndarray] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        if not isinstance(self.m, LDM):
            raise ValueError("`sr` is now only available for `LDM` models")
        factor = 2 ** (len(self.m.first_stage.core.channel_multipliers) - 1)
        res = read_image(image, round(max_wh / factor), anchor=anchor)
        wh_ratio = res.original_size[0] / res.original_size[1]
        zh, zw = res.image.shape[-2:]
        sr_size = (zw, zw / wh_ratio) if zw > zh else (zh * wh_ratio, zh)
        sr_size = tuple(map(lambda n: round(factor * n), sr_size))  # type: ignore
        cond = torch.from_numpy(2.0 * res.image - 1.0).to(self.device)
        z = torch.randn_like(cond)
        return self.sample(
            1,
            export_path,
            z=z,
            original_size=sr_size,
            alpha=res.alpha if alpha is None else alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            verbose=verbose,
            **kwargs,
        )

    def semantic2img(
        self,
        semantic: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 16,
        max_wh: int = 512,
        alpha: Optional[np.ndarray] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        err_fmt = "`{}` is needed for `semantic2img`"
        if self.cond_model is None:
            raise ValueError(err_fmt.format("cond_model"))
        in_channels = getattr(self.cond_model, "in_channels", None)
        if in_channels is None:
            raise ValueError(err_fmt.format("cond_model.in_channels"))
        factor = getattr(self.cond_model, "factor", None)
        if factor is None:
            raise ValueError(err_fmt.format("cond_model.factor"))
        res = read_image(
            semantic,
            max_wh,
            anchor=anchor,
            to_gray=True,
            resample=Image.NEAREST,
            normalize=False,
        )
        cond = torch.from_numpy(res.image).to(torch.long).to(self.device)
        cond = F.one_hot(cond, num_classes=in_channels)[0]
        cond = cond.half() if self.use_half else cond.float()
        cond = cond.permute(0, 3, 1, 2).contiguous()
        cond = self.get_cond(cond)
        z = torch.randn_like(cond)
        return self.sample(
            1,
            export_path,
            z=z,
            original_size=res.original_size,
            alpha=res.alpha if alpha is None else alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            verbose=verbose,
            **kwargs,
        )

    def load_context(self) -> context_error_handler:
        class _(context_error_handler):
            def __init__(self, api: DiffusionAPI):
                self.api = api
                self.m_cond = api.m.condition_model
                api.m.condition_model = api.cond_model

            def __enter__(self) -> nn.Module:
                class Wrapper(nn.Module):
                    def __init__(self, m: nn.Module) -> None:
                        super().__init__()
                        self.core = m

                return Wrapper(self.api.m)

            def _normal_exit(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
                self.api.m.condition_model = self.m_cond

        return _(self)

    @classmethod
    def from_sd(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        return cls.from_pipeline(ldm_sd(), device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_sd_version(
        cls,
        version: str,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        m = ldm_sd_version(version)
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_sd_anime(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        m = ldm_sd_anime()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_sd_inpainting(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        m = ldm_sd_inpainting()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_sd_v2(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        return cls.from_pipeline(
            ldm_sd_v2(),
            device,
            use_amp=use_amp,
            use_half=use_half,
            clip_skip=1,
        )

    @classmethod
    def from_sd_v2_base(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        return cls.from_pipeline(
            ldm_sd_v2_base(),
            device,
            use_amp=use_amp,
            use_half=use_half,
            clip_skip=1,
        )

    @classmethod
    def from_celeba_hq(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        m = ldm_celeba_hq()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_inpainting(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        m = ldm_inpainting()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_sr(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        return cls.from_pipeline(ldm_sr(), device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_semantic(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        m = ldm_semantic()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    def _get_z(self, img: arr_type) -> Tensor:
        img = 2.0 * img - 1.0
        z = img if isinstance(img, Tensor) else torch.from_numpy(img)
        if self.use_half:
            z = z.half()
        z = z.to(self.device)
        z = self.m._preprocess(z, deterministic=True)
        return z

    def _set_seed_and_variations(
        self,
        seed: Optional[int],
        get_noise: Callable[[], Tensor],
        get_new_z: Callable[[Tensor], Tensor],
        variations: Optional[List[Tuple[int, float]]],
        variation_seed: Optional[int],
        variation_strength: Optional[float],
    ) -> Tuple[Tensor, Tensor]:
        if seed is None:
            seed = new_seed()
        seed = seed_everything(seed)
        self.latest_seed = seed
        z_noise = get_noise()
        self.latest_variation_seed = None
        if variations is not None:
            for v_seed, v_weight in variations:
                seed_everything(v_seed)
                v_noise = get_noise()
                z_noise = slerp(v_noise, z_noise, v_weight)
        if variation_strength is not None:
            random.seed()
            if variation_seed is None:
                variation_seed = new_seed()
            variation_seed = seed_everything(variation_seed)
            self.latest_variation_seed = variation_seed
            variation_noise = get_noise()
            z_noise = slerp(variation_noise, z_noise, variation_strength)
        z = get_new_z(z_noise)
        return z, z_noise

    def _q_sample(
        self,
        z: Tensor,
        num_steps: Optional[int],
        fidelity: float,
        seed: int,
        variations: Optional[List[Tuple[int, float]]] = None,
        variation_seed: Optional[int] = None,
        variation_strength: Optional[float] = None,
        **kwargs: Any,
    ) -> Tuple[Tensor, Tensor, Dict[str, Any]]:
        if num_steps is None:
            num_steps = self.sampler.default_steps
        t = min(num_steps, round((1.0 - fidelity) * (num_steps + 1)))
        ts = get_timesteps(t, 1, z.device)
        if isinstance(self.sampler, (DDIMMixin, KSamplerMixin, DPMSolver)):
            kw = shallow_copy_dict(self.sampler.sample_kwargs)
            kw["total_step"] = num_steps
            safe_execute(self.sampler._reset_buffers, kw)
        z, noise = self._set_seed_and_variations(
            seed,
            lambda: torch.randn_like(z),
            lambda noise_: self.sampler.q_sample(z, ts, noise_),
            variations,
            variation_seed,
            variation_strength,
        )
        kwargs["start_step"] = num_steps - t
        return z, noise, kwargs

    def _img2img(
        self,
        z: Tensor,
        export_path: Optional[str] = None,
        *,
        z_ref: Optional[Tensor] = None,
        z_ref_mask: Optional[Tensor] = None,
        original_size: Optional[Tuple[int, int]] = None,
        alpha: Optional[np.ndarray] = None,
        cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        with switch_sampler_context(self, kwargs.get("sampler")):
            z, noise, kwargs = self._q_sample(z, num_steps, **kwargs)
            return self.sample(
                1,
                export_path,
                z=z,
                z_ref=z_ref,
                z_ref_mask=z_ref_mask,
                z_ref_noise=None if z_ref is None else noise,
                original_size=original_size,
                alpha=alpha,
                cond=cond,
                num_steps=num_steps,
                clip_output=clip_output,
                verbose=verbose,
                **kwargs,
            )


def _ldm(
    model: str,
    latent_size: int,
    latent_in_channels: int,
    latent_out_channels: int,
    **kwargs: Any,
) -> DLPipeline:
    kwargs["img_size"] = latent_size
    kwargs["in_channels"] = latent_in_channels
    kwargs["out_channels"] = latent_out_channels
    model_config = kwargs.setdefault("model_config", {})
    first_stage_kw = model_config.setdefault("first_stage_config", {})
    first_stage_kw.setdefault("report", False)
    first_stage_kw.setdefault("pretrained", True)
    first_stage_model_config = first_stage_kw.setdefault("model_config", {})
    use_loss = first_stage_model_config.setdefault("use_loss", False)
    if not use_loss:

        def state_callback(states: tensor_dict_type) -> tensor_dict_type:
            for key in list(states.keys()):
                if key.startswith("core.loss"):
                    states.pop(key)
            return states

        first_stage_kw["pretrained_state_callback"] = state_callback
    return DLZoo.load_pipeline(model, **kwargs)


def ldm(
    latent_size: int = 32,
    latent_in_channels: int = 4,
    latent_out_channels: int = 4,
    **kwargs: Any,
) -> DLPipeline:
    return _ldm(
        "diffusion/ldm",
        latent_size,
        latent_in_channels,
        latent_out_channels,
        **kwargs,
    )


def ldm_vq(
    latent_size: int = 64,
    latent_in_channels: int = 3,
    latent_out_channels: int = 3,
    **kwargs: Any,
) -> DLPipeline:
    return _ldm(
        "diffusion/ldm.vq",
        latent_size,
        latent_in_channels,
        latent_out_channels,
        **kwargs,
    )


def ldm_sd(pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    return _ldm("diffusion/ldm.sd", 64, 4, 4, pretrained=pretrained, **kwargs)


def ldm_sd_version(version: str, pretrained: bool = True) -> DLPipeline:
    return ldm_sd(pretrained, download_name=f"ldm_sd_{version}")


def ldm_sd_anime(pretrained: bool = True) -> DLPipeline:
    return ldm_sd_version("anime_nai", pretrained)


def ldm_sd_inpainting(pretrained: bool = True, **kw: Any) -> DLPipeline:
    return _ldm("diffusion/ldm.sd_inpainting", 64, 9, 4, pretrained=pretrained, **kw)


def ldm_sd_v2(pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    return _ldm("diffusion/ldm.sd_v2", 64, 4, 4, pretrained=pretrained, **kwargs)


def ldm_sd_v2_base(pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    return _ldm("diffusion/ldm.sd_v2_base", 64, 4, 4, pretrained=pretrained, **kwargs)


def ldm_celeba_hq(pretrained: bool = True) -> DLPipeline:
    return ldm_vq(
        pretrained=pretrained,
        download_name="ldm_celeba_hq",
        model_config=dict(
            ema_decay=None,
            first_stage_config=dict(
                pretrained=False,
            ),
        ),
    )


def ldm_inpainting(pretrained: bool = True) -> DLPipeline:
    return ldm_vq(
        pretrained=pretrained,
        latent_in_channels=7,
        download_name="ldm_inpainting",
        model_config=dict(
            ema_decay=None,
            start_channels=256,
            num_heads=8,
            num_head_channels=None,
            resample_with_resblock=True,
            condition_type=CONCAT_TYPE,
            first_stage_config=dict(
                pretrained=False,
                model_config=dict(
                    attention_type="none",
                ),
            ),
        ),
    )


def ldm_sr(pretrained: bool = True) -> DLPipeline:
    return ldm_vq(
        pretrained=pretrained,
        latent_in_channels=6,
        download_name="ldm_sr",
        model_config=dict(
            ema_decay=None,
            start_channels=160,
            attention_downsample_rates=[8, 16],
            channel_multipliers=[1, 2, 2, 4],
            condition_type=CONCAT_TYPE,
            first_stage_config=dict(
                pretrained=False,
            ),
        ),
    )


def ldm_semantic(pretrained: bool = True) -> DLPipeline:
    return ldm_vq(
        pretrained=pretrained,
        latent_size=128,
        latent_in_channels=6,
        download_name="ldm_semantic",
        model_config=dict(
            ema_decay=None,
            start_channels=128,
            num_heads=8,
            num_head_channels=None,
            attention_downsample_rates=[8, 16, 32],
            channel_multipliers=[1, 4, 8],
            condition_type=CONCAT_TYPE,
            condition_model="rescaler",
            condition_config=dict(
                num_stages=2,
                in_channels=182,
                out_channels=3,
            ),
            first_stage_config=dict(
                pretrained=False,
            ),
        ),
    )


__all__ = [
    "DiffusionAPI",
]



import torch

import numpy as np

from .inference.utils import load_model
from .inference.predictor import Predictor
from .....misc.toolkit import download_model


class ImageHarmonizationAPI:
    def __init__(self, device: torch.device) -> None:
        m = load_model("hrnet32_idih256", download_model("hrnet"))
        self.device = device
        self.predictor = Predictor(m, device, with_flip=False)

    def to(self, device: torch.device) -> None:
        self.device = device
        self.predictor.net.to(device)

    # return uint8 image array
    def run(self, image: np.ndarray, mask: np.ndarray) -> np.ndarray:
        return self.predictor.predict(image, mask)


__all__ = [
    "ImageHarmonizationAPI",
]



import cflearn

from typing import Dict
from typing import Union
from cftool.misc import shallow_copy_dict
from cftool.array import tensor_dict_type


def get_open_clip_config(model_name: str) -> Dict[str, Union[str, int]]:
    def _get_num_heads(sub_cfg: Dict[str, int], default: int) -> int:
        num_heads = sub_cfg.get("heads")
        if num_heads is not None:
            return num_heads
        width = sub_cfg["width"]
        head_width = sub_cfg.get("head_width")
        if head_width is not None:
            if width % head_width != 0:
                raise ValueError("`head_width` should be divided by `width`")
            return width // head_width
        return default

    try:
        import open_clip
    except:
        raise ValueError("`open_clip` should be installed for `get_open_clip_config`")
    cfg = open_clip.get_model_config(model_name)
    if cfg is None:
        raise ValueError(f"unrecognized model_name `{model_name}` occurred")
    vision_cfg = cfg["vision_cfg"]
    text_cfg = cfg["text_cfg"]
    return dict(
        latent_dim=cfg.get("embed_dim", 512),
        vision_latent_dim=vision_cfg["width"],
        vision_patch_size=vision_cfg.get("patch_size", 32),
        vision_num_heads=_get_num_heads(vision_cfg, 12),
        vision_num_layers=vision_cfg.get("layers", 12),
        vision_feedforward_activation="GELU",
        text_latent_dim=text_cfg["width"],
        text_num_heads=_get_num_heads(text_cfg, 8),
        text_num_layers=text_cfg.get("layers", 12),
        text_feedforward_activation="GELU",
    )


def convert_open_clip(model_name: str, pretrained: str) -> tensor_dict_type:
    try:
        import open_clip
    except:
        raise ValueError("`open_clip` should be installed for `convert_open_clip`")
    cf_clip_config = get_open_clip_config(model_name)
    oc_model = open_clip.create_model(model_name, pretrained)
    d = shallow_copy_dict(oc_model.state_dict())
    cf_model = cflearn.api.clip_model(model_config=cf_clip_config, pretrained=False)
    md = shallow_copy_dict(cf_model.state_dict())
    extra = {
        "logit_scale": d.pop("logit_scale"),
        "vit.encoder.head_token": d.pop("visual.class_embedding"),
        "vit.encoder.pos_encoding.pos_encoding": d.pop("visual.positional_embedding"),
        "vit.output_projection": d.pop("visual.proj"),
        "token_embedding.weight": d.pop("token_embedding.weight"),
        "text_projection.weight": d.pop("text_projection").t(),
        "text_transformer.encoder.pos_encoding.pos_encoding": d.pop(
            "positional_embedding"
        ),
    }
    re_order_keys = [
        "text_transformer.attention_mask",
    ]
    for k, v in extra.items():
        mv = md.pop(k)
        md[k] = v.view(mv.shape)
    for k in re_order_keys:
        mv = md.pop(k)
        md[k] = mv
    m_keys = list(md)
    for i, (k, v) in enumerate(d.items()):
        mk = m_keys[i]
        mv = md[mk]
        md[mk] = v.view(mv.shape)
    return md



import numpy as np

from enum import Enum
from typing import List
from typing import Tuple
from typing import Union
from typing import NamedTuple
from cftool.misc import print_warning
from cftool.array import is_float
from cftool.array import get_unique_indices


num_split_type = Union[int, float]
xy_type = Tuple[np.ndarray, np.ndarray]


def get_num(n: num_split_type, total: int) -> int:
    if n < 1.0 or (n == 1.0 and isinstance(n, float)):
        return int(round(total * n))
    return int(n)


def preprocess_labels(y: np.ndarray) -> np.ndarray:
    if len(y.shape) == 1:
        y = y[..., None]
    return y


class DataOrder(str, Enum):
    NONE = "none"
    TOP_DOWN = "top_down"
    BOTTOM_UP = "bottom_up"


class DataSplit(NamedTuple):
    x: np.ndarray
    y: np.ndarray
    x_remained: np.ndarray
    y_remained: np.ndarray
    split_indices: np.ndarray
    remained_indices: np.ndarray

    @property
    def xy(self) -> xy_type:
        return self.x, self.y

    @property
    def xy_remained(self) -> xy_type:
        return self.x_remained, self.y_remained


class DataSplitter:
    """
    Util class for dividing dataset based on task type
    * If it's regression task, it's not difficult to split data
    * If it's classification task, we need to split data based on the labels, because we need to:
      * ensure the divided data contain all labels available
      * ensure the label distribution is identical across all divided data.

    Examples
    --------
    >>> import numpy as np
    >>>
    >>> from cflearn.data.ml.utils import DataSplitter
    >>>
    >>> x = np.arange(12).reshape([6, 2])
    >>> # create an imbalance dataset
    >>> y = np.zeros(6, int)
    >>> y[[-1, -2]] = 1
    >>> splitter = DataSplitter()
    >>> # labels in result will keep its ratio
    >>> result = splitter.split(x, y, 3)
    >>> print(result.y.ravel())  # [0 0 1]
    >>> result = splitter.split(x, y, 0.5)
    >>> print(result.y.ravel())  # [0 0 1]
    >>> # at least one sample of each class will be kept
    >>> y[-2] = 0
    >>> result = splitter.split(x, y, 2)
    >>> # [0 0 0 0 0 1] [0 1]
    >>> print(y, result.dataset.y.ravel())

    """

    def __init__(
        self,
        *,
        order: DataOrder = DataOrder.NONE,
        shuffle: bool = True,
    ) -> None:
        if shuffle and order != DataOrder.NONE:
            raise ValueError(
                "`order` should be set to 'none' "
                f"when `shuffle` is set to 'True', but '{order}' found"
            )
        self._order = order
        self._shuffle = shuffle

    # internal

    def _split_with_indices(
        self,
        x: np.ndarray,
        y: np.ndarray,
        split_indices: np.ndarray,
        remained_indices: np.ndarray,
    ) -> DataSplit:
        return DataSplit(
            x[split_indices],
            y[split_indices],
            x[remained_indices],
            y[remained_indices],
            split_indices,
            remained_indices,
        )

    # api

    @property
    def order(self) -> DataOrder:
        return self._order

    @order.setter
    def order(self, value: DataOrder) -> None:
        self._order = value
        if self._shuffle and value != DataOrder.NONE:
            print_warning(
                f"`order` is set to '{value}', "
                "so `shuffle` will be set to 'False' by force"
            )
            self._shuffle = False

    @property
    def shuffle(self) -> bool:
        return self._shuffle

    @shuffle.setter
    def shuffle(self, value: bool) -> None:
        self._shuffle = value
        if value and self._order != DataOrder.NONE:
            print_warning(
                f"`shuffle` is set to 'True', "
                "so `order` will be set to 'none' by force"
            )
            self._order = DataOrder.NONE

    # will treat float `y` as regression datasets and int `y` as classification datasets
    def split(self, x: np.ndarray, y: np.ndarray, n: num_split_type) -> DataSplit:
        method = "reg" if is_float(y) else "clf"
        return getattr(self, f"split_{method}")(x, y, n)

    # split regression datasets
    def split_reg(self, x: np.ndarray, y: np.ndarray, n: num_split_type) -> DataSplit:
        y = preprocess_labels(y)
        total = len(x)
        n = get_num(n, total)
        if self._shuffle:
            base = np.random.permutation(total)
        else:
            base = np.arange(total)
            if self._order == DataOrder.BOTTOM_UP:
                base = np.ascontiguousarray(base[::-1])
        split_indices = base[:n]
        remained_indices = base[n:]
        return self._split_with_indices(x, y, split_indices, remained_indices)

    # split classification datasets
    def split_clf(self, x: np.ndarray, y: np.ndarray, n: num_split_type) -> DataSplit:
        y = preprocess_labels(y)
        total = len(x)
        n = get_num(n, total)
        res = get_unique_indices(y.ravel())
        label_counts = res.unique_cnt
        label_ratios = label_counts.astype(np.float64) / total
        unique_labels = res.unique.astype(int)
        num_unique_labels = len(unique_labels)
        if n < num_unique_labels:
            raise ValueError(
                f"at least {num_unique_labels} samples are required because "
                f"we have {num_unique_labels} unique labels"
            )
        label_split_indices = [indices.astype(int) for indices in res.split_indices]
        if num_unique_labels == 1:
            raise ValueError(
                "only 1 unique label is detected, "
                "which is invalid in classification task"
            )
        if self._shuffle:
            tuple(map(np.random.shuffle, label_split_indices))
        elif self._order == DataOrder.BOTTOM_UP:
            label_split_indices = [indices[::-1] for indices in label_split_indices]
            label_split_indices = list(map(np.ascontiguousarray, label_split_indices))
        rounded = np.round(n * label_ratios).astype(int)
        num_samples_per_label = np.maximum(1, rounded)
        # -num_unique_labels <= num_samples_exceeded <= num_unique_labels
        num_samples_exceeded = num_samples_per_label.sum() - n
        # adjust num_samples_per_label to make sure `n` samples are split out
        if num_samples_exceeded != 0:
            sign = np.sign(num_samples_exceeded)
            num_samples_exceeded = abs(num_samples_exceeded)
            arange = np.arange(num_unique_labels)
            chosen_indices = arange[num_samples_per_label != 1]
            np.random.shuffle(chosen_indices)
            num_chosen_indices = len(chosen_indices)
            num_tile = int(np.ceil(num_samples_exceeded / num_chosen_indices))
            num_proceeded = 0
            for _ in range(num_tile - 1):
                num_samples_per_label[chosen_indices] -= sign
                num_proceeded += num_chosen_indices
            for idx in chosen_indices[: num_samples_exceeded - num_proceeded]:
                num_samples_per_label[idx] -= sign
        assert num_samples_per_label.sum() == n
        num_overlap = 0
        split_indices_list: List[np.ndarray] = []
        remained_indices_list: List[np.ndarray] = []
        for indices, num_sample_per_label in zip(
            label_split_indices,
            num_samples_per_label,
        ):
            num_samples_available = len(indices)
            split_indices_list.append(indices[:num_sample_per_label])
            if num_sample_per_label >= num_samples_available:
                num_overlap += num_sample_per_label
                remained_indices_list.append(indices)
            else:
                remained_indices_list.append(indices[num_sample_per_label:])
        split_indices = np.hstack(split_indices_list)
        remained_indices = np.hstack(remained_indices_list)
        base = np.zeros(total)
        base[split_indices] += 1
        base[remained_indices] += 1
        assert np.sum(base >= 2) <= num_overlap
        return self._split_with_indices(x, y, split_indices, remained_indices)


__all__ = [
    "DataSplitter",
]



import torch

import torch.nn as nn

from torch import Tensor
from typing import List
from typing import Union
from typing import Optional

from ....register import register_module
from ...schemas.cv import ImageTranslatorMixin
from ....misc.toolkit import interpolate
from ....modules.blocks import get_clones
from ....modules.blocks import get_conv_blocks
from ....modules.blocks import Conv2d
from ....modules.blocks import Interpolate


class ConvSeq(nn.Sequential):
    def __init__(self, in_channels: int = 3, out_channels: int = 3, dilation: int = 1):
        super().__init__(
            *get_conv_blocks(
                in_channels,
                out_channels,
                3,
                1,
                norm_type="batch",
                activation=nn.ReLU(inplace=True),
                padding=dilation,
                dilation=dilation,
            )
        )


class UNetBase(nn.Module):
    in_block: nn.Module
    down_blocks: nn.ModuleList
    last_down: nn.Module
    up_blocks: nn.ModuleList
    return_up_nets: bool = False

    def forward(
        self,
        net: Tensor,
        *,
        determinate: bool = False,
    ) -> Union[Tensor, List[Tensor]]:
        net = in_net = self.in_block(net)
        down_nets = []
        for down_block in self.down_blocks:
            net = down_block(net)
            down_nets.append(net)
        inner = self.last_down(net)
        up_nets = [inner]
        for down_net, up_block in zip(down_nets[::-1], self.up_blocks):
            if inner.shape != down_net.shape:
                inner = interpolate(
                    inner,
                    anchor=down_net,
                    mode="bilinear",
                    determinate=determinate,
                )
            inner = torch.cat([inner, down_net], dim=1)
            if not isinstance(up_block, nn.ModuleList):
                inner = up_block(inner)
            else:
                for block in up_block:
                    kw = {}
                    if isinstance(block, Interpolate):
                        kw["determinate"] = determinate
                    inner = block(inner, **kw)
            up_nets.append(inner)
        if self.return_up_nets:
            return up_nets
        return inner + in_net


class UNetRS(UNetBase):
    def __init__(
        self,
        in_channels: int = 3,
        mid_channels: int = 12,
        out_channels: int = 3,
        *,
        num_layers: int,
        inner_upsample_mode: str = "bilinear",
        inner_upsample_factor: Optional[int] = None,
    ):
        super().__init__()
        self.msg = f"{in_channels}, {mid_channels}, {out_channels}, {num_layers}"
        self.in_block = ConvSeq(in_channels, out_channels, dilation=1)
        blocks = [ConvSeq(out_channels, mid_channels, dilation=1)]
        for _ in range(num_layers - 2):
            blocks.append(
                nn.Sequential(
                    nn.MaxPool2d(2, stride=2),
                    ConvSeq(mid_channels, mid_channels, dilation=1),
                )
            )
        self.down_blocks = nn.ModuleList(blocks)
        self.last_down = ConvSeq(mid_channels, mid_channels, dilation=2)
        basic_block = ConvSeq(mid_channels * 2, mid_channels, dilation=1)
        if inner_upsample_factor is not None:
            basic_block = nn.ModuleList(
                [
                    basic_block,
                    Interpolate(inner_upsample_factor, inner_upsample_mode),
                ]
            )
        blocks = get_clones(basic_block, num_layers - 2, return_list=True)
        blocks.append(ConvSeq(mid_channels * 2, out_channels, dilation=1))
        self.up_blocks = nn.ModuleList(blocks)

    def extra_repr(self) -> str:
        return self.msg


class UNetFRS(UNetBase):
    def __init__(
        self,
        in_channels: int = 3,
        mid_channels: int = 12,
        out_channels: int = 3,
        *,
        num_layers: int,
    ):
        super().__init__()
        self.msg = f"{in_channels}, {mid_channels}, {out_channels}, {num_layers}"
        self.in_block = ConvSeq(in_channels, out_channels, dilation=1)
        blocks = [ConvSeq(out_channels, mid_channels, dilation=1)]
        for i in range(num_layers - 2):
            blocks.append(ConvSeq(mid_channels, mid_channels, dilation=2 ** (i + 1)))
        self.down_blocks = nn.ModuleList(blocks)
        self.last_down = ConvSeq(mid_channels, mid_channels, dilation=8)
        blocks = []
        for i in range(num_layers - 2):
            dilation = 2 ** (num_layers - i - 2)
            blocks.append(ConvSeq(mid_channels * 2, mid_channels, dilation=dilation))
        blocks.append(ConvSeq(mid_channels * 2, out_channels, dilation=1))
        self.up_blocks = nn.ModuleList(blocks)

    def extra_repr(self) -> str:
        return self.msg


@register_module("u2net")
class U2Net(UNetBase, ImageTranslatorMixin):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        *,
        upsample_mode: str = "bilinear",
        latent_channels: int = 32,
        num_layers: int = 5,
        num_inner_layers: int = 7,
        lite: bool = False,
    ):
        super().__init__()
        self.return_up_nets = True

        mid_nc = latent_channels // 2 if lite else latent_channels
        out_nc = latent_channels * 2
        in_nc = out_nc if lite else latent_channels
        current_layers = num_inner_layers

        self.in_block = nn.Identity()
        blocks = [
            UNetRS(
                in_channels,
                mid_nc,
                out_nc,
                num_layers=current_layers,
                inner_upsample_mode=upsample_mode,
                inner_upsample_factor=2,
            ),
        ]
        for i in range(num_layers - 2):
            if lite:
                in_nc = out_nc
            else:
                in_nc = out_nc
                out_nc *= 2
            current_layers = num_inner_layers - i - 1
            blocks.append(
                nn.Sequential(
                    nn.MaxPool2d(2, stride=2),
                    UNetRS(
                        in_nc,
                        mid_nc,
                        out_nc,
                        num_layers=current_layers,
                        inner_upsample_mode=upsample_mode,
                        inner_upsample_factor=2,
                    ),
                )
            )
            if not lite:
                mid_nc *= 2
        if not lite:
            in_nc *= 2
        blocks.append(
            nn.Sequential(
                nn.MaxPool2d(2, stride=2),
                UNetFRS(in_nc, mid_nc, out_nc, num_layers=current_layers),
            )
        )
        self.down_blocks = nn.ModuleList(blocks)
        self.last_down = nn.Sequential(
            nn.MaxPool2d(2, stride=2),
            UNetFRS(in_nc, mid_nc, out_nc, num_layers=current_layers),
        )
        in_nc *= 2
        ncs = [out_nc] * 2
        blocks = [UNetFRS(in_nc, mid_nc, out_nc, num_layers=current_layers)]
        for i in range(num_layers - 2):
            if not lite:
                mid_nc //= 2
                out_nc //= 2
            blocks.append(
                UNetRS(
                    in_nc,
                    mid_nc,
                    out_nc,
                    num_layers=current_layers,
                    inner_upsample_mode=upsample_mode,
                    inner_upsample_factor=2,
                )
            )
            current_layers += 1
            if not lite:
                in_nc //= 2
            ncs.append(out_nc)
        if not lite:
            mid_nc //= 2
        blocks.append(
            UNetRS(
                in_nc,
                mid_nc,
                out_nc,
                num_layers=current_layers,
                inner_upsample_mode=upsample_mode,
                inner_upsample_factor=2,
            )
        )
        ncs.append(out_nc)
        self.up_blocks = nn.ModuleList(blocks)

        blocks = []
        for i, nc in enumerate(ncs[::-1]):
            blocks.append(Conv2d(nc, out_channels, kernel_size=3, padding=1))
        self.side_blocks = nn.ModuleList(blocks)
        self.out = Conv2d((len(ncs)) * out_channels, out_channels, kernel_size=1)

    def forward(self, net: Tensor, *, determinate: bool = False) -> List[Tensor]:
        up_nets = super().forward(net)
        side_nets: List[Tensor] = []
        for up_net, side_block in zip(up_nets[::-1], self.side_blocks):
            side_net = side_block(up_net)
            if side_nets and side_net.shape != side_nets[0].shape:
                side_net = interpolate(
                    side_net,
                    anchor=side_nets[0],
                    mode="bilinear",
                    determinate=determinate,
                )
            side_nets.append(side_net)
        side_nets.insert(0, self.out(torch.cat(side_nets, dim=1)))
        return side_nets


__all__ = ["U2Net"]



from typing import List
from typing import Union
from typing import Optional

from .monitors import BasicMonitor
from .callbacks import _LogMetricsMsgCallback
from ..schema import TqdmSettings
from ..schema import OptimizerPack
from ..schema import TrainerMonitor
from ..schema import _IMetric
from ..schema import _MultipleMetrics
from ..schema import TrainerConfig
from ..schema import TrainerCallback
from ..trainer import Trainer


def make_trainer(config: TrainerConfig) -> Trainer:
    state_config = config.state_config
    log_steps = config.log_steps
    if state_config is None:
        state_config = {}
    if log_steps is not None:
        state_config.setdefault("num_step_per_log", log_steps)
        state_config.setdefault("snapshot_start_step", log_steps)
        state_config.setdefault("num_step_per_snapshot", log_steps)
    # metrics
    metric_names = config.metric_names
    metric_configs = config.metric_configs
    metric_weights = config.metric_weights
    metrics: Optional[Union[_IMetric, _MultipleMetrics]]
    if metric_names is None:
        metrics = None
    else:
        metrics = _IMetric.fuse(
            metric_names,
            metric_configs,
            metric_weights=metric_weights,
        )
    # monitor
    monitor_names = config.monitor_names
    monitor_configs = config.monitor_configs
    monitors: Optional[List[TrainerMonitor]]
    if monitor_names is None:
        monitors = None
    else:
        monitors = TrainerMonitor.make_multiple(monitor_names, monitor_configs)
    # callback
    callback_names = config.callback_names
    callback_configs = config.callback_configs
    callbacks: Optional[List[TrainerCallback]]
    if callback_names is None:
        callbacks = None
    else:
        callbacks = TrainerCallback.make_multiple(callback_names, callback_configs)
    # optimizer
    optimizer_settings = config.optimizer_settings
    if optimizer_settings is None:
        optimizer_packs = None
    else:
        optimizer_packs = []
        for key, settings in optimizer_settings.items():
            optimizer = settings.get("optimizer")
            if optimizer is None:
                raise ValueError(f"optimizer must be provided (key={key})")
            optimizer_packs.append(
                OptimizerPack(
                    key,
                    optimizer,
                    settings.get("scheduler"),
                    settings.get("optimizer_config"),
                    settings.get("scheduler_config"),
                )
            )
    # tqdm
    tqdm_settings = config.tqdm_settings
    if tqdm_settings is None:
        tqdm_settings = {}
    use_tqdm = tqdm_settings.setdefault("use_tqdm", False)
    # epoch
    max_epoch = config.max_epoch
    num_epoch = config.num_epoch
    fixed_epoch = config.fixed_epoch
    if fixed_epoch is not None:
        num_epoch = max_epoch = fixed_epoch
    if max_epoch < num_epoch:
        raise ValueError("`max_epoch` should not be smaller than `num_epoch`")
    # default behaviors
    if monitors is None:
        monitors = [BasicMonitor()]
    if callbacks is None:
        callbacks = [_LogMetricsMsgCallback(not use_tqdm)]
    return Trainer(
        state_config,
        num_epoch=num_epoch,
        max_epoch=max_epoch,
        fixed_steps=config.fixed_steps,
        valid_portion=config.valid_portion,
        amp=config.amp,
        clip_norm=config.clip_norm,
        metrics=metrics,
        use_losses_as_metrics=config.use_losses_as_metrics,
        loss_metrics_weights=config.loss_metrics_weights,
        recompute_train_losses_in_eval=config.recompute_train_losses_in_eval,
        monitors=monitors,
        callbacks=callbacks,
        lr=config.lr,
        optimizer_name=config.optimizer_name,
        scheduler_name=config.scheduler_name,
        optimizer_config=config.optimizer_config,
        scheduler_config=config.scheduler_config,
        update_scheduler_per_epoch=config.update_scheduler_per_epoch,
        optimizer_packs=optimizer_packs,
        use_zero=config.use_zero,
        workplace=config.workplace,
        data_info_name=config.data_info_name,
        metrics_log_file=config.metrics_log_file,
        finetune_config=config.finetune_config,
        tqdm_settings=TqdmSettings(
            use_tqdm,
            tqdm_settings.setdefault("use_step_tqdm", use_tqdm),
            tqdm_settings.setdefault("use_tqdm_in_validation", False),
            tqdm_settings.setdefault("in_distributed", False),
            tqdm_settings.setdefault("tqdm_position", 0),
            tqdm_settings.setdefault("tqdm_desc", "epoch"),
        ),
    )


__all__ = [
    "make_trainer",
]



import os
import json
import math

import numpy as np

from abc import abstractmethod
from tqdm import tqdm
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import Protocol
from typing import NamedTuple
from collections import Counter
from dataclasses import dataclass
from cftool.dist import Parallel
from cftool.misc import grouped
from cftool.misc import hash_code
from cftool.misc import print_info
from cftool.misc import print_warning
from cftool.misc import DataClassBase
from cftool.array import get_unique_indices
from cftool.array import SharedArray
from cftool.array import StrideArray

from .basic import register_ml_data
from .basic import register_ml_data_processor
from .basic import IMLData
from .basic import IMLBatch
from .basic import MLDatasetTag
from .basic import IMLDataProcessor
from .basic import IMLPreProcessedData


TIME_SERIES_NAME = "_internal.ts"


@dataclass
class TimeSeriesConfig(DataClassBase):
    gap: int
    x_window: int
    y_window: int
    id_column: int
    time_columns: List[int]
    # `validation_split` will be included in the validation dataset
    validation_split: Optional[int] = None
    # `validation_end` will NOT be included in the validation dataset
    validation_end: Optional[int] = None
    # `test_split` will be included in the test dataset
    test_split: Optional[int] = None
    # `test_end` will NOT be included in the test dataset
    test_end: Optional[int] = None
    num_test: Optional[int] = None
    enforce_num_test: bool = False
    enforce_test_valid: bool = False
    num_workers: Optional[int] = None
    random_sample_ratio: Optional[float] = None
    for_inference: bool = False
    sanity_check: bool = False
    no_cache: bool = False
    verbose: bool = True
    cache_folder: str = "cache"

    @property
    def span(self) -> int:
        if self.for_inference:
            return self.x_window
        return self.x_window + self.y_window + self.gap

    @property
    def hash(self) -> str:
        d = self.asdict()
        for k in self.hash_excludes:
            d.pop(k)
        return hash_code(str({k: d[k] for k in sorted(d)}))

    @property
    def hash_excludes(self) -> List[str]:
        excludes = [
            "num_workers",
            "random_sample_ratio",
            "sanity_check",
            "no_cache",
            "verbose",
            "cache_folder",
        ]
        if not self.for_inference:
            excludes += [
                "test_split",
                "test_end",
                "num_test",
                "enforce_num_test",
                "enforce_test_valid",
            ]
        return excludes

    @property
    def use_validation(self) -> bool:
        return not self.for_inference and self.validation_split is not None


@dataclass
class TimeSeriesCachePaths(DataClassBase):
    data_path: str
    merged_indices_path: str
    split_indices_path: str
    valid_indices_path: str

    @property
    def folder(self) -> str:
        return os.path.dirname(self.data_path)

    @property
    def all_exists(self) -> bool:
        return all(map(os.path.isfile, self.asdict().values()))


class TimeSeriesDataBundle(NamedTuple):
    config: TimeSeriesConfig
    # all (flattened) data, should already be sorted by time
    # shape: [N, d]
    data: np.ndarray
    # first hierarchy: sorted ids
    # second hierarchy: indices of each id that point to the id's data
    split_indices: List[List[int]]
    # all rolled indices (including those which contain different ids)
    # shape: [N - span + 1, span]
    rolled_indices: np.ndarray
    # indices of the `rolled_indices` which are 'valid'
    # 1. len(valid_indices) == number of valid samples
    # 2. sample procedure:
    #      i = randint(n)
    #   -> idx = valid_indices[i]
    #   -> indices = rolled_indices[idx]
    #   -> sample = data[indices]
    valid_indices: np.ndarray

    def __len__(self) -> int:
        return len(self.valid_indices)

    def fetch_batch(self, indices: Union[int, List[int], np.ndarray]) -> IMLBatch:
        indices_mat = self.rolled_indices[self.valid_indices[indices]]
        if isinstance(indices, int) or np.isscalar(indices):
            data_batch = self.data[indices_mat]
            x_batch = data_batch[: self.config.x_window]
        else:
            shape = [len(indices), self.config.span, -1]
            data_batch = self.data[indices_mat.ravel()].reshape(shape)
            x_batch = data_batch[:, : self.config.x_window]
        if self.config.for_inference:
            y_batch = None
        else:
            y_batch = data_batch[:, self.config.x_window + self.config.gap :]
        return IMLBatch(
            x_batch,
            None if y_batch is None else y_batch,
            others=dict(full=data_batch),
        )

    def to_loader(
        self,
        *,
        batch_size: Optional[int] = None,
        tqdm_desc: Optional[str] = None,
    ) -> Iterable[IMLBatch]:
        n = len(self)
        it: Iterable = range(n)
        if batch_size is not None:
            n = math.ceil(n / batch_size)
            it = iter(list(i) for i in grouped(it, batch_size, keep_tail=True))
        if tqdm_desc is not None:
            it = tqdm(it, desc=tqdm_desc, total=n)
        return iter(self.fetch_batch(i) for i in it)


class ITimeSeriesProcessor(IMLDataProcessor):
    config: TimeSeriesConfig

    _train_bundle: TimeSeriesDataBundle
    _validation_bundle: Optional[TimeSeriesDataBundle]

    config_base = TimeSeriesConfig
    cache_paths_base = TimeSeriesCachePaths

    # abstract

    @property
    @abstractmethod
    def tag(self) -> str:
        pass

    @classmethod
    @abstractmethod
    def prepare_data(cls, config: TimeSeriesConfig) -> np.ndarray:
        pass

    @abstractmethod
    def get_time_anchors(self, times: np.ndarray) -> np.ndarray:
        """
        return the time anchors based on `data[time_columns]`, which will be used to
        calculate the sorted indices
        """

    @abstractmethod
    def check_valid(self, mat: np.ndarray) -> bool:
        """check whether `mat` is a valid data sample"""

    def sanity_check(self, tag: MLDatasetTag, bundle: TimeSeriesDataBundle) -> None:
        def check_ids(data: np.ndarray) -> np.ndarray:
            ids_ = data[..., self.config.id_column].astype(int)
            if np.abs(ids_[:, 1:] - ids_[:, :-1]).sum().item() != 0:
                raise ValueError(f"multiple ids occurred in one sample : {ids_}")
            return ids_

        def get_time_anchors(data: np.ndarray) -> np.ndarray:
            times = data[..., self.config.time_columns]
            time_anchors = self.get_time_anchors(times.reshape([-1, times.shape[-1]]))
            time_anchors = time_anchors.reshape(times.shape[:-1])
            return time_anchors

        min_max_x_anchor = None
        min_max_y_anchor = None
        max_max_x_anchor = None
        max_max_y_anchor = None
        id_counts: Counter = Counter()
        loader = bundle.to_loader(batch_size=512, tqdm_desc="check_split")
        for batch in loader:
            x, y = batch.input, batch.labels
            full = x if y is None or batch.others is None else batch.others["full"]
            ids = check_ids(full)
            x_anchors = get_time_anchors(x)
            y_anchors = None if y is None else get_time_anchors(y)
            max_x_anchors = x_anchors.max(axis=-1)
            max_y_anchors = None if y_anchors is None else y_anchors.max(axis=-1)
            if tag == MLDatasetTag.TRAIN:
                if not self.config.for_inference:
                    if self.config.validation_split is not None:
                        anchors = x_anchors if y_anchors is None else y_anchors
                        max_anchors = (
                            max_x_anchors if max_y_anchors is None else max_y_anchors
                        )
                        if np.any(max_anchors >= self.config.validation_split):
                            msg = f"train data exceeds validation split : {anchors}"
                            raise ValueError(msg)
                else:
                    if self.config.num_test is not None:
                        for id_ in ids[:, 0]:
                            if id_counts[id_] >= self.config.num_test:
                                raise ValueError("test data exceeds num_test")
                            id_counts[id_] += 1
                    elif self.config.test_split is not None:
                        if np.any(max_x_anchors < self.config.test_split):
                            msg = f"test data exceeds test split : {x_anchors}"
                            raise ValueError(msg)
                        if self.config.test_end is not None and np.any(
                            max_x_anchors >= self.config.test_end
                        ):
                            msg = f"test data exceeds test end : {x_anchors}"
                            raise ValueError(msg)
            elif tag == MLDatasetTag.VALID:
                if self.config.validation_split is not None and np.any(
                    max_x_anchors < self.config.validation_split
                ):
                    raise ValueError(
                        f"validation data exceeds validation split : {x_anchors}"
                    )
                if self.config.validation_end is not None:
                    anchors = x_anchors if y_anchors is None else y_anchors
                    max_anchors = (
                        max_x_anchors if max_y_anchors is None else max_y_anchors
                    )
                    if np.any(max_anchors >= self.config.validation_end):
                        msg = f"validation data exceeds validation end : {anchors}"
                        raise ValueError(msg)
            # update statics
            if min_max_x_anchor is None:
                min_max_x_anchor = max_x_anchors.min().item()
            else:
                min_max_x_anchor = min(max_x_anchors.min().item(), min_max_x_anchor)
            if max_max_x_anchor is None:
                max_max_x_anchor = max_x_anchors.max().item()
            else:
                max_max_x_anchor = max(max_x_anchors.max().item(), max_max_x_anchor)
            if max_y_anchors is not None:
                if min_max_y_anchor is None:
                    min_max_y_anchor = max_y_anchors.min().item()
                else:
                    min_max_y_anchor = min(max_y_anchors.min().item(), min_max_y_anchor)
            if max_y_anchors is not None:
                if max_max_y_anchor is None:
                    max_max_y_anchor = max_y_anchors.max().item()
                else:
                    max_max_y_anchor = max(max_y_anchors.max().item(), max_max_y_anchor)
        print(f"> min max_x_anchor of {tag} : {min_max_x_anchor}")
        print(f"> min max_y_anchor of {tag} : {min_max_y_anchor}")
        print(f"> max max_x_anchor of {tag} : {max_max_x_anchor}")
        print(f"> max max_y_anchor of {tag} : {max_max_y_anchor}")

    # utils

    def verbose(self, msg: str) -> None:
        if self.config.verbose:
            print_info(msg)

    @property
    def hashed_cache_folder(self) -> str:
        return os.path.join(self.config.cache_folder, self.tag, self.config.hash)

    def get_cache_paths(self, tag: MLDatasetTag) -> TimeSeriesCachePaths:
        folder = os.path.join(self.hashed_cache_folder, tag)
        paths = dict(
            data_path=os.path.join(folder, "data.npy"),
            merged_indices_path=os.path.join(folder, "merged_indices.npy"),
            split_indices_path=os.path.join(folder, "split_indices.json"),
            valid_indices_path=os.path.join(folder, "valid_indices.npy"),
        )
        paths = {k: os.path.abspath(v) for k, v in paths.items()}
        return TimeSeriesCachePaths(**paths)

    def load_cache_paths(self, paths: TimeSeriesCachePaths) -> TimeSeriesDataBundle:
        self.verbose("loading data")
        data = np.load(paths.data_path)
        self.verbose("loading rolled indices")
        np_merged_indices = np.load(paths.merged_indices_path)
        rolled_indices = StrideArray(np_merged_indices).roll(self.config.span)
        self.verbose("loading split indices")
        with open(paths.split_indices_path, "r") as f:
            split_indices = json.load(f)
        self.verbose("loading valid indices")
        valid_indices = np.load(paths.valid_indices_path)
        # return
        args = self.config.copy(), data, split_indices, rolled_indices, valid_indices
        return TimeSeriesDataBundle(*args)

    def _get_valid_indices(
        self,
        i: int,
        num: int,
        data: np.ndarray,
        rolled_indices: np.ndarray,
        num_indices_cumsum: np.ndarray,
        boolean_mask: Optional[np.ndarray],
    ) -> Optional[List[int]]:
        if num < self.config.span:
            return [] if boolean_mask is None else None
        cumulate = 0 if i == 0 else num_indices_cumsum[i - 1]
        valid_indices: Optional[List[int]] = [] if boolean_mask is None else None
        for j in range(num - self.config.span + 1):
            idx = j + cumulate
            is_valid = False
            if self.config.num_test is not None and self.config.enforce_test_valid:
                is_valid = True
            elif self.check_valid(data[rolled_indices[idx]]):
                is_valid = True
            if is_valid:
                if boolean_mask is not None:
                    boolean_mask[idx] = True
                elif valid_indices is not None:
                    valid_indices.append(idx)
        return valid_indices

    def get_bundle(
        self,
        x_train: np.ndarray,
        indices: np.ndarray,
        cache_paths: TimeSeriesCachePaths,
    ) -> TimeSeriesDataBundle:
        # data
        data = x_train[indices]
        if not self.config.no_cache:
            np.save(cache_paths.data_path, data)
        # split indices
        self.verbose("group by id")
        ids = data[..., self.config.id_column]
        split_indices = get_unique_indices(ids).split_indices
        split_indices = [indices.tolist() for indices in split_indices]
        if self.config.num_test is not None and self.config.enforce_num_test:
            span = self.config.x_window + self.config.num_test - 1
            split_indices = [indices[-span:] for indices in split_indices]
        if not self.config.no_cache:
            with open(cache_paths.split_indices_path, "w") as f:
                json.dump(split_indices, f)
        # rolled indices
        merged_indices = []
        for indices in split_indices:
            merged_indices += indices
        np_merged_indices = np.array(merged_indices, int)
        rolled_indices = StrideArray(np_merged_indices).roll(self.config.span)
        if not self.config.no_cache:
            np.save(cache_paths.merged_indices_path, np_merged_indices)
        # valid indices
        fn = self._get_valid_indices
        num_indices = np.array(list(map(len, split_indices)), int)
        num_indices_cumsum = np.cumsum(num_indices)
        if not self.config.num_workers:
            valid_indices_list: List[int] = []
            for i, num in enumerate(tqdm(num_indices, desc="check_valid")):
                i_args = i, num, data, rolled_indices, num_indices_cumsum, None
                valid_indices_list += fn(*i_args) or []
            valid_indices = np.array(valid_indices_list, int)
        else:
            self.verbose("preparing shared arrays")
            shared_arrays = (
                SharedArray.from_data(data),
                SharedArray.from_data(np_merged_indices),
                SharedArray.from_data(num_indices_cumsum),
                SharedArray(np.bool, [num_indices_cumsum[-1]]),
            )
            common_args = [shared.value for shared in shared_arrays]
            common_args[1] = StrideArray(common_args[1]).roll(self.config.span)
            args_list: List[List[Any]] = [[] for _ in range(6)]
            for i, num in enumerate(num_indices):
                for j, arg in enumerate([i, num, *common_args]):
                    args_list[j].append(arg)
            self.verbose("launching workers")
            Parallel(self.config.num_workers).grouped(fn, *args_list)
            valid_indices = np.nonzero(common_args[-1])[0]
            for shared in shared_arrays:
                shared.destroy()
        if not self.config.no_cache:
            np.save(cache_paths.valid_indices_path, valid_indices)
        # random sample
        random_ratio = self.config.random_sample_ratio
        if not self.config.for_inference and random_ratio is not None:
            n = len(valid_indices)
            sampled = np.random.permutation(n)[: round(n * random_ratio)]
            valid_indices = valid_indices[sampled]
        # return
        args = self.config.copy(), data, split_indices, rolled_indices, valid_indices
        return TimeSeriesDataBundle(*args)

    def get_split(
        self,
        split: int,
        end: Optional[int],
        sorted_indices: np.ndarray,
        sorted_anchors: np.ndarray,
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        # `split` belongs to `right_indices`
        unique_anchors = np.unique(sorted_anchors)
        right_mask = unique_anchors >= split
        right_end_mask = None
        if end is not None:
            right_end_mask = unique_anchors < end
            right_mask = right_mask & right_end_mask
        y_span = self.config.gap + self.config.y_window
        if right_mask.sum().item() <= y_span:
            if self.config.verbose:
                print_warning(
                    "validation data is not enough when "
                    f"`validation_split` is set to `{split}`"
                    f"& `validation_end` is set to `{end}`"
                )
            return sorted_indices, None
        left_idx = np.argmax(right_mask).item() - 1
        left_anchor = unique_anchors[left_idx]
        left_mask = sorted_anchors <= left_anchor
        right_idx = max(0, left_idx - self.config.x_window + 2)
        right_anchor = unique_anchors[right_idx]
        right_mask = sorted_anchors >= right_anchor
        if right_end_mask is not None:
            right_end_idx = np.argmin(right_end_mask) - 1
            right_end_idx = min(len(unique_anchors) - 1, right_end_idx.item())
            right_end_anchor = unique_anchors[right_end_idx]
            right_mask = right_mask & (sorted_anchors <= right_end_anchor)
        left_indices = sorted_indices[left_mask]
        right_indices = sorted_indices[right_mask]
        return left_indices, right_indices

    # inheritance

    def build_with(self, config: TimeSeriesConfig, x_train: np.ndarray) -> None:  # type: ignore
        self.config = config
        if not config.no_cache:
            os.makedirs(self.hashed_cache_folder, exist_ok=True)
            with open(os.path.join(self.hashed_cache_folder, "config.json"), "w") as f:
                json.dump(config.asdict(), f)
        # cache
        tr_paths = self.get_cache_paths(MLDatasetTag.TRAIN)
        cv_paths = self.get_cache_paths(MLDatasetTag.VALID)
        if not config.no_cache:
            os.makedirs(tr_paths.folder, exist_ok=True)
            if config.use_validation:
                os.makedirs(cv_paths.folder, exist_ok=True)
        # load
        if (
            not config.no_cache
            and tr_paths.all_exists
            and (not config.use_validation or cv_paths.all_exists)
        ):
            self.verbose("loading train caches")
            self._train_bundle = self.load_cache_paths(tr_paths)
            if not config.use_validation:
                self._validation_bundle = None
            else:
                self.verbose("loading validation caches")
                self._validation_bundle = self.load_cache_paths(cv_paths)
        # build
        else:
            self.verbose("sort by time")
            time_columns = x_train[..., config.time_columns].astype(int)
            time_anchors = self.get_time_anchors(time_columns).ravel()
            sorted_indices = np.argsort(time_anchors)
            if config.use_validation:
                sorted_anchors = time_anchors[sorted_indices]
                tr_indices, cv_indices = self.get_split(
                    config.validation_split,  # type: ignore
                    config.validation_end,
                    sorted_indices,
                    sorted_anchors,
                )
                self.verbose("generating train bundle")
                train_bundle = self.get_bundle(x_train, tr_indices, tr_paths)
                if cv_indices is None:
                    validation_bundle = None
                else:
                    self.verbose("generating validation bundle")
                    validation_bundle = self.get_bundle(x_train, cv_indices, cv_paths)
                self._train_bundle = train_bundle
                self._validation_bundle = validation_bundle
            else:
                tag = "test" if config.for_inference else "train"
                self.verbose(f"generating {tag} bundle")
                if not config.for_inference:
                    indices = sorted_indices
                else:
                    if config.num_test is not None and config.test_end is not None:
                        msg = "should not set `test_end` when `num_test` is set"
                        raise ValueError(msg)
                    if config.num_test is None and config.test_split is None:
                        indices = sorted_indices
                    elif config.num_test is not None and config.enforce_num_test:
                        indices = sorted_indices
                    else:
                        sorted_anchors = time_anchors[sorted_indices]
                        if config.test_split is not None:
                            test_split = config.test_split
                        else:
                            test_split = sorted_anchors[-config.num_test]  # type: ignore
                        _, indices = self.get_split(
                            test_split,
                            config.test_end,
                            sorted_indices,
                            sorted_anchors,
                        )
                self._train_bundle = self.get_bundle(x_train, indices, tr_paths)
                self._validation_bundle = None
        self.verbose("done")

    def preprocess(self, config: TimeSeriesConfig) -> IMLPreProcessedData:  # type: ignore
        num_test_enforced = config.num_test is not None and config.enforce_test_valid
        if config.sanity_check and not num_test_enforced:
            self.sanity_check(MLDatasetTag.TRAIN, self._train_bundle)
            if self._validation_bundle is not None:
                self.sanity_check(MLDatasetTag.VALID, self._validation_bundle)
        if self._validation_bundle is None:
            x_valid = None
        else:
            x_valid = self._validation_bundle.data
        return IMLPreProcessedData(
            self._train_bundle.data,
            x_valid=x_valid,
            num_history=self.config.x_window,
            is_classification=False,
        )

    def dumps(self) -> Dict[str, Any]:
        d = self.config.asdict()
        d["tr_cache_paths"] = self.get_cache_paths(MLDatasetTag.TRAIN).asdict()
        if not self.config.use_validation:
            d["cv_cache_paths"] = None
        else:
            d["cv_cache_paths"] = self.get_cache_paths(MLDatasetTag.VALID).asdict()
        return d

    def loads(self, dumped: Dict[str, Any]) -> None:
        def _load(d: Dict[str, Any], tag: str) -> TimeSeriesDataBundle:
            self.verbose(f"loading {tag} caches")
            cache_paths = self.cache_paths_base(**d)
            return self.load_cache_paths(cache_paths)

        tr_paths_d = dumped.pop("tr_cache_paths")
        cv_paths_d = dumped.pop("cv_cache_paths")
        self.config = self.config_base(**dumped)
        self.verbose("load begin")
        self._train_bundle = _load(tr_paths_d, "train")
        if cv_paths_d is None:
            self._validation_bundle = None
        else:
            self._validation_bundle = _load(cv_paths_d, "validation")
        self.verbose("load complete")

    # callbacks

    def get_num_samples(self, x: np.ndarray, tag: MLDatasetTag) -> int:
        if tag == MLDatasetTag.TRAIN:
            return len(self._train_bundle)
        if self._validation_bundle is None:
            raise ValueError(
                "`_validation_bundle` is not ready "
                "but `get_num_samples` with `tag=valid` is called"
            )
        return len(self._validation_bundle)

    def fetch_batch(
        self,
        x: np.ndarray,
        y: Optional[np.ndarray],
        indices: Union[int, List[int], np.ndarray],
        tag: MLDatasetTag,
    ) -> IMLBatch:
        if tag == MLDatasetTag.TRAIN:
            bundle = self._train_bundle
        else:
            if self._validation_bundle is None:
                raise ValueError(
                    "`_validation_bundle` is not ready "
                    "but `fetch_batch` with `tag=valid` is called"
                )
            bundle = self._validation_bundle
        return bundle.fetch_batch(indices)

    # api

    def get_flatten(self, indices: np.ndarray, tag: MLDatasetTag) -> np.ndarray:
        if tag == MLDatasetTag.TRAIN:
            bundle = self._train_bundle
        else:
            if self._validation_bundle is None:
                raise ValueError(
                    "`_validation_bundle` is not ready "
                    "but `get_flatten` with `tag=valid` is called"
                )
            bundle = self._validation_bundle
        indices = bundle.rolled_indices[bundle.valid_indices[indices]][..., 0]
        return bundle.data[indices]


@register_ml_data(TIME_SERIES_NAME)
class TimeSeriesData(IMLData):
    processor: ITimeSeriesProcessor
    processor_type = TIME_SERIES_NAME

    def __init__(
        self,
        ts_config: TimeSeriesConfig,
        *,
        data: Optional[np.ndarray] = None,
        processor: Optional[IMLDataProcessor] = None,
        shuffle_train: bool = True,
        shuffle_valid: bool = False,
        batch_size: int = 128,
        valid_batch_size: int = 512,
        use_numpy: bool = False,
    ):
        self.ts_config = ts_config
        if data is None:
            processor_cls = ITimeSeriesProcessor.get(self.processor_type)
            data = processor_cls.prepare_data(ts_config)
        super().__init__(
            data,
            processor=processor,
            shuffle_train=shuffle_train,
            shuffle_valid=shuffle_valid,
            batch_size=batch_size,
            valid_batch_size=valid_batch_size,
            use_numpy=use_numpy,
            for_inference=ts_config.for_inference,
        )

    @property
    def train_bundle(self) -> TimeSeriesDataBundle:
        return self.processor._train_bundle

    @property
    def validation_bundle(self) -> Optional[TimeSeriesDataBundle]:
        return self.processor._validation_bundle

    @property
    def processor_build_config(self) -> TimeSeriesConfig:
        return self.ts_config

    @property
    def processor_preprocess_config(self) -> TimeSeriesConfig:
        return self.ts_config


def make_ts_test_data(
    config: TimeSeriesConfig,
    *,
    data: Optional[np.ndarray] = None,
    test_split: Optional[int] = None,
    test_end: Optional[int] = None,
    num_test: Optional[int] = None,
    enforce_num_test: Optional[bool] = None,
    enforce_test_valid: Optional[bool] = None,
    sanity_check: bool = True,
    use_numpy: bool = False,
    no_cache: bool = True,
    batch_size: int = 1024,
) -> TimeSeriesData:
    config = config.copy()
    config.for_inference = True
    config.validation_split = None
    config.validation_end = None
    if test_split is not None:
        config.test_split = test_split
    if test_end is not None:
        config.test_end = test_end
    if num_test is not None:
        config.num_test = num_test
    if enforce_num_test is not None:
        config.enforce_num_test = enforce_num_test
    if enforce_test_valid is not None:
        config.enforce_test_valid = enforce_test_valid
    config.sanity_check = sanity_check
    config.no_cache = no_cache
    config.y_window = 0
    config.gap = 0
    test_data = TimeSeriesData(
        config,
        data=data,
        batch_size=batch_size,
        shuffle_train=False,
        use_numpy=use_numpy,
    )
    test_data.prepare(None)
    return test_data


class ITSRollingInferencePostProcess(Protocol):
    def __call__(self, i: int, predictions: np.ndarray, raw_data: np.ndarray) -> None:
        """we need to fill `predictions` to `raw_data` to perform rolling inference"""


def ts_rolling_inference(
    m: Any,
    data: np.ndarray,
    config: TimeSeriesConfig,
    steps: int,
    *,
    num_test: Optional[int] = None,
    enforce_num_test: bool = False,
    sanity_check: bool = True,
    use_numpy: bool = False,
    no_cache: bool = True,
    batch_size: int = 1024,
    postprocess: Optional[ITSRollingInferencePostProcess] = None,
) -> np.ndarray:
    results = None
    for i in range(steps):
        print_info(f"rolling inference {i+1} / {steps}")
        if results is None:
            test_arr = data
        else:
            test_arr = np.vstack([data, results])
        test_data = make_ts_test_data(
            config.copy(),
            data=test_arr,
            num_test=num_test,
            enforce_num_test=enforce_num_test,
            sanity_check=sanity_check,
            use_numpy=use_numpy,
            no_cache=no_cache,
            batch_size=batch_size,
        )
        i_predictions = m.predict(test_data)["predictions"]
        i_raw_data = (
            test_data.processor.get_flatten(
                np.array(list(range(len(i_predictions)))),
                MLDatasetTag.TRAIN,
            )
            .copy()
            .astype(np.float64)
        )
        if postprocess is not None:
            postprocess(i, i_predictions, i_raw_data)
        print_info(f"filled data ({i+1}) {i_raw_data.shape}")
        if results is None:
            results = i_raw_data
        else:
            results = np.vstack([results, i_raw_data])
    return results


def register_ts_processor() -> Callable:
    return register_ml_data_processor(TIME_SERIES_NAME)


__all__ = [
    "make_ts_test_data",
    "ts_rolling_inference",
    "register_ts_processor",
    "TimeSeriesConfig",
    "ITimeSeriesProcessor",
    "TimeSeriesData",
]
